﻿; *---------------------------------*
; *   OpenGL Base Example           *
; *   xorc1zt                       *
; *   2011                          *
; *---------------------------------*

EnableExplicit
IncludeFile "glbase.pb"

Global quitmainloop.b = #False
Global viewmatrix.MATRIX
Global projmatrix.MATRIX
Global program_id.i
Global texture_id.i

;- Declarations
Declare tickWindow()
Declare onResize( height.i, width.i )
Declare onRender()
Declare onInit()

;- Start
pbglOpenWindow( 100, 100, 800, 600, "Hello OpenGL" )
onInit()

;- Main loop
Repeat
    onRender()
    FlipBuffers()
    tickWindow()
Until quitmainloop

;- Procedures
Procedure tickWindow()
    Select WindowEvent()
        Case #PB_Event_CloseWindow
            quitmainloop = #True
        Case #PB_Event_SizeWindow
            onResize( WindowWidth( 0 ),  WindowHeight( 0 ) )
    EndSelect
EndProcedure

Procedure onResize( width.i,  height.i)
    Debug "New size! : "+ Str( width )+"x"+Str( height )
    glViewport( 0, 0, width, height )
    projmatrix\perspective( 70, 1, 100, width/height )
EndProcedure

Procedure onRender()
    Define vertex_location.i
    Define color_location.i
    Static angle.f
    
    glClearScreen( 0.2, 0.2, 0.2 )

    angle + 1.0
    vertex_location = glGetAttribLocation( program_id, "in_Vertex" )
    glVertexAttribPointer( vertex_location, 3, #GL_FLOAT, #GL_FALSE, 0, ?cubevertex )
    glEnableVertexAttribArray( vertex_location )
    
    color_location = glGetAttribLocation( program_id, "in_Color" )
    glVertexAttribPointer( color_location, 3, #GL_FLOAT, #GL_FALSE, 0, ?cubecolor )
    glEnableVertexAttribArray( color_location )
    
    glUseProgram( program_id )
    glUniformMatrix4fv( glGetUniformLocation( program_id, "projection" ), 1, #GL_FALSE, projmatrix\getAddress() )
    
    viewmatrix\identity()
    viewmatrix\lookAt( 0, 3, 8, 0, 0, 0, 0, 1, 0 )
    
    ;cube 1 at {0,0,0}
    viewmatrix\push()
    viewmatrix\rotateZ( angle )
    glUniformMatrix4fv( glGetUniformLocation( program_id, "modelview" ), 1, #GL_FALSE, viewmatrix\getAddress() ) 
    glDrawArrays( #GL_TRIANGLES, 0, 36 )
    viewmatrix\pop()
    
    ;cube 2 at {-3,0,0}
    viewmatrix\push()
    viewmatrix\translate( -3, 0, 0 )
    viewmatrix\rotateY( angle )
    glUniformMatrix4fv( glGetUniformLocation( program_id, "modelview" ), 1, #GL_FALSE, viewmatrix\getAddress() ) 
    glDrawArrays( #GL_TRIANGLES, 0, 36 )
    viewmatrix\pop()
    
    ;cube 3 at {3,0,0}
    viewmatrix\push()
    viewmatrix\translate( 3, 0, 0 )
    viewmatrix\rotateX( angle )
    glUniformMatrix4fv( glGetUniformLocation( program_id, "modelview" ), 1, #GL_FALSE, viewmatrix\getAddress() ) 
    glDrawArrays( #GL_TRIANGLES, 0, 36 )
    viewmatrix\pop()
    
    glUseProgram( 0 )
    
    glDisableVertexAttribArray( vertex_location )
    glDisableVertexAttribArray( color_location )
    
EndProcedure

Procedure onInit()
    Define vertex_shader.s
    Define fragment_shader.s
    
    ;initialize our matrix
    viewmatrix = createMatrix( "viewmatrix" )
    projmatrix = createMatrix( "projmatrix" )
    
    ;vertex shader source code
    vertex_shader = "#version 330"+#CRLF$
    vertex_shader + "in vec3 in_Vertex;"+#CRLF$
    vertex_shader + "in vec3 in_Color;"+#CRLF$
    vertex_shader + "out vec3 color;"+#CRLF$
    vertex_shader + "uniform mat4 projection;"+#CRLF$
    vertex_shader + "uniform mat4 modelview;"+#CRLF$
    vertex_shader + "void main()"+#CRLF$
    vertex_shader + "{"+#CRLF$
    vertex_shader + "gl_Position = projection * modelview * vec4(in_Vertex, 1.0);"+#CRLF$
    vertex_shader + "color = in_Color;"+#CRLF$
    vertex_shader + "}"
    
    ;fragment shader source code
    fragment_shader = "#version 330"+#CRLF$
    fragment_shader + "in vec3 color;"+#CRLF$
    fragment_shader + "out vec4 out_Color;"+#CRLF$
    fragment_shader + "void main()"+#CRLF$
    fragment_shader + "{"+#CRLF$
    fragment_shader + "out_Color = vec4(color, 1.0);"+#CRLF$
    fragment_shader + "}"
    
    ;create a program with our shaders
    program_id = pbglCreateProgramFast( vertex_shader, fragment_shader )
EndProcedure

DataSection
    cubecolor:
    cubevertex:
    Data.f  1.0, 1.0,-1.0; north
    Data.f -1.0,-1.0,-1.0
    Data.f -1.0, 1.0,-1.0
    Data.f  1.0, 1.0,-1.0
    Data.f  1.0,-1.0,-1.0
    Data.f -1.0,-1.0,-1.0
    Data.f -1.0,-1.0,-1.0; west
    Data.f -1.0,-1.0, 1.0  
    Data.f -1.0, 1.0, 1.0    
    Data.f -1.0,-1.0,-1.0
    Data.f -1.0, 1.0, 1.0
    Data.f -1.0, 1.0,-1.0
    Data.f  1.0, 1.0, 1.0; east
    Data.f  1.0,-1.0,-1.0
    Data.f  1.0, 1.0,-1.0
    Data.f  1.0,-1.0,-1.0
    Data.f  1.0, 1.0, 1.0
    Data.f  1.0,-1.0, 1.0 
    Data.f -1.0, 1.0, 1.0; south
    Data.f -1.0,-1.0, 1.0
    Data.f  1.0,-1.0, 1.0 
    Data.f  1.0, 1.0, 1.0
    Data.f -1.0, 1.0, 1.0
    Data.f  1.0,-1.0, 1.0
    Data.f  1.0, 1.0, 1.0; top
    Data.f  1.0, 1.0,-1.0
    Data.f -1.0, 1.0,-1.0    
    Data.f  1.0, 1.0, 1.0
    Data.f -1.0, 1.0,-1.0
    Data.f -1.0, 1.0, 1.0
    Data.f  1.0,-1.0, 1.0; bot
    Data.f -1.0,-1.0, 1.0
    Data.f -1.0,-1.0,-1.0
    Data.f -1.0,-1.0,-1.0
    Data.f  1.0,-1.0,-1.0
    Data.f  1.0,-1.0, 1.0
    
    
EndDataSection